"use strict";

var STAGE_W = 400; // ステージのサイズ
var STAGE_H = 640;
var INTERVAL = 30; // タイマー周期
var FRICTION = 0.99; // 摩擦係数

var BALL_N = 10;   // ボールの数
var BALL_R = 4;    // ボールの半径
var BALL_IX = STAGE_W / 2; // ボールの初期座標
var BALL_IY = STAGE_H - BALL_R - 1;
var BALL_IV_MAX = 120; // ボール初速の最大値

var HOLE_MIN = 1;  // 穴の数の最小・最大値
var HOLE_MAX = 9;
var HOLE_MASS = 2000; // 穴の質量
var HOLE_R = 8;    // 穴の半径

var GOAL_MIN = 1;  // ゴール数の最小・最大値
var GOAL_MAX = 5;
var GOAL_W = 16;   // ゴールのサイズ
var GOAL_H = 8;

var balls;         // ボールの配列 {x, y, vx, vy}
var hole_n = 5;    // 穴の数
var holes;         // 穴の配列 {x, y}
var goal_n = 3;    // ゴールの数
var goals_x;       // ゴールのX座標
var point = 0;     // 得点
var total = 0;     // ボールの数

var cv;            // ボール描画用canvas
var overlay_cv;    // オーバーレイcanvas
var $ball_img = null; // ボールの画像(jQueryオブジェクト)

var drag_x, drag_y;   // ドラッグ座標
var dragging = false; // ドラッグフラグ

/**
    2点A-B間のユークリッド距離を求める
    @param ax    点AのX座標
    @param ay    点AのY座標
    @param bx    点BのX座標
    @param by    点BのY座標
    @return 点A-B間のユークリッド距離
 */
function distance(ax, ay, bx, by) {
    return Math.sqrt((ax - bx) * (ax - bx) + (ay - by) * (ay - by));
}

/**
    背景の画像を合成し、#bg_imgのsrcとする。
    描画用canvasとして、一時的にoverlay_cvを用いる。
 */
function draw_stage() {
    var x, y, i, j, r, v, img, s;
    img = overlay_cv.createImageData(STAGE_W, STAGE_H);
    j = 0; // 画素インデックス
    // 各点でのポテンシャルを求める。
    for (y = 0; y < STAGE_H; y++) {
        for (x = 0; x < STAGE_W; x++) {
            v = 0;
            for (i = 0; i < hole_n; i++) {
                s = holes[i];
                r = distance(s.x, s.y, x, y);
                if (r <= HOLE_R) {
                    v = 255;
                    break;
                }
                v += 10 * s.m / r / r;
            }
            v = 128 - v;
            if (v < 0) { v = 0; }
            if (v > 100 && Math.random() > 0.999) {
                // たまに背景に星を描画
                v = Math.random() * 128 + 64;
                img.data[j++] = img.data[j++] = v;
                img.data[j++] = v / 2;
            } else { 
                // 穴の周囲はポテンシャルを濃淡で描画
                img.data[j++] = img.data[j++] = v / 2;
                img.data[j++] = v;
            }
            img.data[j++] = 255; // α値
        }
    }
    // canvasにポテンシャルを描画
    overlay_cv.putImageData(img, 0, 0);

    // ゴールを描画
    for (i = 0; i < goals_x.length; i++) {
        x = goals_x[i];
        overlay_cv.fillStyle = "#000000";
        overlay_cv.fillRect(x - GOAL_W, 0, GOAL_W * 2, GOAL_H / 2);
        overlay_cv.beginPath();
        overlay_cv.moveTo(x - GOAL_H / 2, GOAL_H);
        overlay_cv.lineTo(x + GOAL_H / 2, GOAL_H);
        overlay_cv.lineTo(x, 0);
        overlay_cv.closePath();
        overlay_cv.fillStyle = "#FF0000";
        overlay_cv.fill();
    }

    // canvasからimgに画像を転送。
    $('#bg_img')[0].src = $('#overlay_cv')[0].toDataURL('image/png');
    overlay_cv.clearRect(0, 0, STAGE_W, STAGE_H);
}

/**
    ボールにかかる重力加速度を計算する
    @param x ボールのX座標
    @param y ボールのY座標
    @return 落下したら false、スイングバイなら重力加速度
 */
function calc_acceleration(x, y) {
    var i, ax = 0, ay = 0, mr2, s, r;
    for (i = 0; i < hole_n; i++) {
        s = holes[i];
        r = distance(s.x, s.y, x, y);
        if (r <= HOLE_R) { return false; }
        mr2 = s.m / r / r;
        ax += (s.x - x) / r * mr2;
        ay += (s.y - y) / r * mr2;
    }
    return {x: ax, y: ay};
}

/**
    ルンゲ・クッタ法により、ボールの位置と速度を計算する
    @param ball ボールオブジェクト {x,y,vx,vy}
    @return    false 落下もしくはゴール、true 生存
 */
function calc_ball_position(ball) {
    var i, a, a2, a3, a4,
        b2x, b2y, v2x, v2y, b3x, b3y, v3x, v3y, b4x, b4y, v4x, v4y;

    a = calc_acceleration(ball.x, ball.y);
    if (!a) { return false; }

    b2x = ball.x + ball.vx / 2;
    b2y = ball.y + ball.vy / 2;
    v2x = ball.vx + a.x / 2;
    v2y = ball.vy + a.y / 2;
    a2 = calc_acceleration(b2x, b2y);
    if (!a2) { return false; }

    b3x = ball.x + v2x / 2;
    b3y = ball.y + v2y / 2;
    v3x = ball.vx + a2.x / 2;
    v3y = ball.vy + a2.y / 2;
    a3 = calc_acceleration(b3x, b3y);
    if (!a3) { return false; }

    b4x = ball.x + v3x;
    b4y = ball.y + v3y;
    v4x = ball.vx + a3.x;
    v4y = ball.vy + a3.y;
    a4 = calc_acceleration(b4x, b4y);
    if (!a4) { return false; }

    ball.x += (ball.vx + v2x * 2 + v3x * 2 + v4x) / 6;
    ball.y += (ball.vy + v2y * 2 + v3y * 2 + v4y) / 6;
    ball.vx += (a.x + a2.x * 2 + a3.x * 2 + a4.x) / 6;
    ball.vy += (a.y + a2.y * 2 + a3.y * 2 + a4.y) / 6;

    // 摩擦による速度低下
    ball.vx = FRICTION * ball.vx;
    ball.vy = FRICTION * ball.vy;

    // 壁の跳ね返り
    if (ball.y < BALL_R) { // 上
        // ゴールに到達か？
        for (i = 0; i < goals_x.length; i++) {
            if (ball.x >= goals_x[i] - GOAL_W &&
                ball.x < goals_x[i] + GOAL_W) {
                point++;
                $('#point').val(point);
                return false;
            }
        }
        ball.y = 2 * BALL_R - ball.y;
        ball.vy = -ball.vy;
    } else if (ball.y >= STAGE_H - BALL_R) { // 下
        ball.y = 2 * STAGE_H - 2 * BALL_R -    ball.y;
        ball.vy = -ball.vy;
    }
    if (ball.x < BALL_R) { // 左
        ball.x = 2 * BALL_R - ball.x;
        ball.vx = -ball.vx;
    } else if (ball.x >= STAGE_W - BALL_R) { // 右
        ball.x = 2 * STAGE_W - 2 * BALL_R - ball.x;
        ball.vx = -ball.vx;
    }

    return true;
}

/**
    ベクトルvを基準ベクトルdに正射影
    @param vx    ベクトルvのX成分
    @param vy    ベクトルvのY成分
    @param dx    基準ベクトルのX成分
    @param dy    基準ベクトルのY成分
    @return ベクトルvを基準ベクトルdに正射影したベクトル {x,y}
 */
function projection(vx, vy, dx, dy) {
    var a = (vx * dx + vy * dy) / (dx * dx + dy * dy);
    return { x: a * dx, y: a * dy };
}

/**
    衝突判定。
    ballsの要素を総当たりで調べて、衝突していたら反発させる。
 */
function check_collision() {
    var i, j, bi, bj, n, dx, dy, via, vib, vja, vjb;
    n = balls.length;
    for (i = 0; i < n - 1; i++) {
        bi = balls[i];
        for (j = i + 1; j < n; j++) {
            bj = balls[j];
            if (distance(bi.x, bi.y, bj.x, bj.y) < 2 * BALL_R) {
                // 衝突。運動量を入れ替え
                dx = bi.x - bj.x;
                dy = bi.y - bj.y;
                if (dx !== 0 || dy !== 0) {
                    via = projection(bi.vx, bi.vy, dx, dy);
                    vib = projection(bi.vx, bi.vy, dy, -dx);
                    vja = projection(bj.vx, bj.vy, dx, dy);
                    vjb = projection(bj.vx, bj.vy, dy, -dx);
                    bi.vx = vja.x + vib.x;
                    bi.vy = vja.y + vib.y;
                    bj.vx = via.x + vjb.x;
                    bj.vy = via.y + vjb.y;
                }
            }
        }
    }
}

/**
    入力値を最小値・最大値に収まるようにする。
    @param v 入力値
    @param min 最小値
    @param max 最大値
    @return min..maxに収めた値
 */
function limit_value(v, min, max) {
    if (v < min) { return min; }
    if (v > max) { return max; }
    return v;
}

/**
    ステージを初期化する。
    #hole_n と #goal_n の値を使って穴とゴールを再配置する。
 */
function init_stage() {
    var i;

    // 穴をランダムに生成
    hole_n = $('#hole_n').spinner('value');
    hole_n = limit_value(hole_n, HOLE_MIN, HOLE_MAX);
    $('#hole_n').spinner('value', hole_n);
    holes = [];
    for (i = 0; i < hole_n; i++) {
        holes[i] = {
            x: Math.random() * STAGE_W * 0.8 + STAGE_W * 0.1,
            y: STAGE_H * 0.1 * (2 + 7.0 * i / hole_n),
            m: Math.random() * HOLE_MASS + HOLE_MASS / 4
        };
    }

    // ゴールを配置
    goal_n = $('#goal_n').spinner('value');
    goal_n = limit_value(goal_n, GOAL_MIN, GOAL_MAX);
    $('#goal_n').spinner('value', goal_n);
    goals_x = [];
    for (i = 0; i < goal_n; i++) {
        goals_x[i] = STAGE_W * i / goal_n + STAGE_W / goal_n / 2;
    }
}

/**
    タイマーイベントハンドラ。INTERVAL後に再度呼び出される。
    ボールを移動する。
 */
function on_tick() {
    var i, b, nballs;

    cv.clearRect(0, 0, STAGE_W, STAGE_H);

    // balls のうち、移動後に生存したボールをnballsに記録する
    nballs = [];
    for (i = 0; i < balls.length; i++) {
        b = balls[i];
        if (calc_ball_position(b)) {
            cv.drawImage($ball_img[0], (b.x - BALL_R)>>0, (b.y - BALL_R)>>0);
            nballs.push(b);
        }
    }
    balls = nballs;  // 落下したボールはここで参照が切れて解放

    check_collision();

    // 弾切れなら、ボールを描画しない
    $ball_img.css('display', (balls.length < BALL_N ? 'inline' : 'none'));

    window.setTimeout('on_tick()', INTERVAL);
}

/**
    マウス押下イベントハンドラ
    ステージ内で、弾切れでなければ、発射準備する
    @param e イベントオブジェクト
 */
function on_mousedown(e) {
    if (e.pageX >= STAGE_W || e.pageY >= STAGE_H) { return; }
    e.preventDefault();
    e.stopPropagation();

    if (balls.length < BALL_N) {
        dragging = true;
        on_mousemove(e);
    }
}

/**
    マウス解放イベントハンドラ
    発射準備中なら、ボールを初期化して追加する。
 */
function on_mouseup(e) {
    if (!dragging) { return; }
    dragging = false;
    e.preventDefault();
    e.stopPropagation();

    balls.push({ x: BALL_IX, y: BALL_IY,
                 vx: (drag_x - BALL_IX) / 10,
                 vy: (drag_y - BALL_IY) / 10 });

    overlay_cv.clearRect(0, 0, STAGE_W, STAGE_H);
    total++;
    $('#total').val(total);
}

/**
    マウス移動イベントハンドラ
    発射準備中なら、ボールの発射速度を描画する。
    発射速度は、BALL_IV_MAX以下にする。
 */
function on_mousemove(e) {
    var d;
    if (!dragging) { return; }
    e.preventDefault();
    e.stopPropagation();

    drag_x = e.pageX;
    drag_y = e.pageY;
    d = distance(drag_x, drag_y, BALL_IX, BALL_IY);
    if (d > BALL_IV_MAX) {
        drag_x = BALL_IX + BALL_IV_MAX * (drag_x - BALL_IX) / d;
        drag_y = BALL_IY + BALL_IV_MAX * (drag_y - BALL_IY) / d;
    }
    // 発射速度を描画
    overlay_cv.clearRect(0, 0, STAGE_W, STAGE_H);
    overlay_cv.beginPath();
    overlay_cv.moveTo(BALL_IX, BALL_IY);
    overlay_cv.lineTo(drag_x, drag_y);
    overlay_cv.strokeStyle = '#FFFFFF';
    overlay_cv.lineWidth = 1;
    overlay_cv.stroke();
}

/**
    ステージを再構成
 */
function on_reset() {
    balls = [];
    point = total = 0;
    $('#point').val(point);
    $('#total').val(total);
    init_stage();
    draw_stage();
}

// ページを初期化
$(document).ready(function() {
    // canvasの取得要素
    cv = $('#cv')[0].getContext('2d');
    overlay_cv = $('#overlay_cv')[0].getContext('2d');
    // ボール画像要素の取得
    $ball_img = $('#ball')
    $ball_img.css({ left: (STAGE_W / 2 - BALL_R)>>0 + 'px', 
                    top: (STAGE_H - BALL_R - 1 - BALL_R) + 'px'} );

    // イベントハンドラの設定
    $(document).mousedown(on_mousedown)
               .mouseup(on_mouseup)
               .mousemove(on_mousemove);
    $('#hole_n').spinner({ min:HOLE_MIN, max: HOLE_MAX, stop: on_reset})
                .val(hole_n);
    $('#goal_n').spinner({ min:GOAL_MIN, max: GOAL_MAX, stop: on_reset})
                .val(goal_n);
    $('#reset').button().click(on_reset);

    on_reset();
    on_tick();
});
